/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <libintl.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "lib_mpgheader.h"
#include "lib_id3v1.h"
#include "lib_id3v2.h"
#include "lib_vorbis.h"
#include "engine_fileinfo.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

GHashTable *fileinfo_hash = NULL;

/***************************************************************************************
 * STATICS
 ***************************************************************************************/
/*
 * Initialize a new FileInfo struct.
 * Returns: A pointer to the new struct.
 */
static FileInfo *fileinfo_new(const gchar * path)
{
  FileInfo *info = g_new0(FileInfo, 1);
  strncpy(info->filename, path, 2047);
  return(info);
}


/*
 * Free a fileinfo struct.
 */
static void fileinfo_destroy(FileInfo *info)
{
  if (info->header)
    g_free(info->header);
  if (info->tag[0])
    g_free(info->tag[0]);
  if (info->tag[1])
    g_free(info->tag[1]);
  if (info->summary)
    g_free(info->summary);
  g_free(info);
}


/*
 * Create a short summary of header and tag informations.
 */
static gchar *mp3_summary(FileInfo *info)
{
  GString *summary = g_string_sized_new(1024);
  const gchar *mode = NULL;
  gchar *ret = NULL;
  gint mins, secs;
  
  if (info->header) {
    // Version:
    g_string_sprintf(summary,
         _("MPEG V%s, Layer %i\nSample rate: %i\nBitrate: %i\n"),
         info->header->version, info->header->layer,
         info->header->frequency, info->header->bitrate);
  
    // Time:
    mins = info->header->seconds / 60;
    secs = info->header->seconds % 60;
  
    g_string_sprintfa(summary, _("Time: %i:%02i\n"), mins, secs);
  
    // Mode:
    switch (info->header->mode) {
    case 0:
      mode = _("Stereo");
      break;
    case 1:
      mode = _("Joint Stereo");
      break;
    case 2:
      mode = _("2 Channel");
      break;
    case 3:
      mode = _("Mono");
      break;
    default:
      g_assert_not_reached();
    }
    
    g_string_sprintfa(summary, _("Mode: %s\n"), mode);
    
    if (info->header->copyright)
      g_string_append(summary, _("Copyright present\n"));
  }

  if (info->tag_type == ID3 && info->tag[0])
    g_string_append(summary, _("ID3V1 Tag present\n"));
  
  if (info->tag_type == OGG && info->tag[0])
    g_string_append(summary, _("OGG Tag present\n"));
  
  if (info->tag[1])
    g_string_append(summary, _("ID3V2 Tag present\n"));
  
  ret = summary->str;
  g_string_free(summary, FALSE);
  
  /* Remove the last newline character. */
  *(ret + strlen(ret) - 1) = '\0';
  
  return(ret);
}


/*
 * Return a duplicate of a tag.
 */
static Tag *tagdup(Tag *tag)
{
  Tag *dup = g_new0(Tag, 1);
  memcpy(dup, tag, sizeof(Tag));
  return(dup);
}


/* --------------------------------- Hash Functions ---------------------------------- */
/*
 * Return the address of the hashtable. If none exists yet, create a new hashtable.
 */
static GHashTable *fileinfo_hash_get(void)
{
  // use a double hash, hashing file name to inode, then inode
  // to mp3info. This complicates things, because if there are a
  // couple of different filenames pointing to an inode, what do
  // we display for a file? or do we just return the same
  // mp3info struct and take the file name out of it altogether?
  // but then how to update it?

  if (fileinfo_hash == NULL)
    fileinfo_hash = g_hash_table_new(g_str_hash, g_str_equal);

  return(fileinfo_hash);
}


/*
 * The function needed to free the whole hash with its values.
 */
static GHRFunc free_data(gchar * path, FileInfo * info, gpointer trash)
{
  fileinfo_destroy(info);
  return(NULL);
}
/* ----------------------------- End Hash Functions ---------------------------------- */
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Get the header and tag informations of a file.
 * If initialize is TRUE, the information is updated,
 * otherwise we get it from the memory (hash lookup).
 */
FileInfo *fileinfo_get(const char *path, short int initialize)
{
  GHashTable *store = fileinfo_hash_get();
  FileInfo *info = NULL;
  struct stat trash;
  
  info = g_hash_table_lookup(store, path);
  
  if (!info) {
    info = fileinfo_new(path);
    g_hash_table_insert(store, info->filename, info);
  }
  
  if (initialize || !info->initialized)
    fileinfo_read_1(path, info);

  return(info);
}


/*
 * Read the header and tag informations of a file.
 */
int fileinfo_read_1(const char *filename, FileInfo *info)
{
  struct stat filestatus;
  struct passwd *pwd;
  struct group  *grp;
  Tag tag;
  
  g_assert(info != NULL);
  
  /* Free all data attached. */
  if (info->tag[0]) {
    g_free(info->tag[0]);
    info->tag[0] = NULL;
  }
  if (info->tag[1]) {
    g_free(info->tag[1]);
    info->tag[1] = NULL;
  }
  if (info->header) {
    g_free(info->header);
    info->header = NULL;
  }
  if (info->summary) {
    g_free(info->header);
    info->summary = NULL;
  }
  
  if (access(filename, R_OK) != 0)
    return(-1);
  
  /* Wipe out the fileinfo. */
  strncpy(info->filename, filename, 2047);
  memset(((gchar *)info) + 2048, 0, sizeof(FileInfo) - 2048);
  
  if (stat(filename, &filestatus) != 0)
    return(-2);
  
  /* Get filesize */
  info->filesize = (gulong)filestatus.st_size;
  
  /* Get owner */
  pwd = getpwuid(filestatus.st_uid);
  if (pwd && pwd->pw_name)
    strncpy(info->owner, pwd->pw_name, 511);
  else
    strncpy(info->owner, "?", 511);
  
  /* Group */
  grp = getgrgid(filestatus.st_gid);
  if (grp && grp->gr_name)
    strncpy(info->group, grp->gr_name, 511);
  else
    strncpy(info->group, "?", 511);
  
  /* Mode */
  sprintf(info->mode, "%u%u%u",
    (gint)(filestatus.st_mode & 0700) >> 6,
    (gint)(filestatus.st_mode & 0070) >> 3,
    (gint)(filestatus.st_mode & 0007));
  
  /* Read the tags. */
  if (get_id3v1_tag(&tag, filename) == 0) {
    info->tag[0] = tagdup(&tag);
    info->tag_type = ID3;
  }
  if (get_id3v2_tag(&tag, filename) == 0) {
    info->tag[1] = tagdup(&tag);
    info->tag_type = ID3;
  }
#ifdef ENABLE_OGG
  if (get_vorbis_tag(&tag, filename) == 0) {
    info->tag[0] = tagdup(&tag);
    info->tag_type = OGG;
  }
#endif
  
  /* Get mpg header infos. */
  info->header = malloc(sizeof(MpgHeader));
  if (get_mpgheader(info->header, filename) != 0) {
    g_free(info->header);
    info->header = NULL;
  }
  
  info->summary = mp3_summary(info);
  
  if (*info->status == '\0')
    strncpy(info->status, _("Fresh"), 255);
  
  info->initialized = TRUE;
  info->updated = TRUE;
  
  return(0);
}


/*
 * Destroy the whole hash and all file informations.
 */
void fileinfo_destroy_all(void)
{
  if (fileinfo_hash == NULL)
    return;
  
  g_hash_table_foreach_remove(fileinfo_hash, (GHRFunc)free_data, NULL);
  g_hash_table_destroy(fileinfo_hash);
  fileinfo_hash = NULL;
}
