/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <libintl.h>
#include <sys/stat.h>
#include "engine_init.h"
#include "engine_fileinfo.h"
#include "engine_conditions.h"
#include "engine_rules.h"
#include "engine_ruleset.h"
#include "engine_jobs.h"
#include "lib_id3v1.h"
#include "lib_id3v2.h"
#include "lib_vorbis.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern Engine engine;
extern Cfg cfg;

/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static Job *job_new(FileInfo *info)
{
  Job *job = (Job *)calloc(1, sizeof(Job));
  
  job->info = info;
  
  job->destinfo = (FileInfo *)malloc(sizeof(FileInfo));
  memcpy(job->destinfo, info, sizeof(FileInfo));
  
  job->destinfo->tag[0] = (Tag *)calloc(1, sizeof(Tag));
  if (info->tag[0])
    memcpy(job->destinfo->tag[0], info->tag[0], sizeof(Tag));
//  else
  //  printf("NO V1!!!!\n");
  job->destinfo->tag[1] = (Tag *)calloc(1, sizeof(Tag));
  if (info->tag[1])
    memcpy(job->destinfo->tag[1], info->tag[1], sizeof(Tag));
//  else
  //  printf("NO V2!!!!\n");
  
  job->updated = TRUE;
  
  return(job);
}


/*
 * Destroy the "job" struct and all appended data EXECEPT the fileinfo.
 */
static void job_destroy(Job *job)
{
  if (!job)
    return;
  if (job->ruleset)
    ruleset_unref(job->ruleset);
  if (job->destinfo && job->destinfo->tag[0])
    free(job->destinfo->tag[0]);
  if (job->destinfo && job->destinfo->tag[1])
    free(job->destinfo->tag[1]);
  if (job->destinfo)
    free(job->destinfo);
  free(job);
}


/*
 * Change the job so that the "preview" is changed to FALSE.
 */
static void job_preview_make_nonpreview(Job *job)
{
  g_assert(job != NULL);
  job->preview = FALSE;
}


void file_rename(Job *job)
{
  struct stat trash;
  
  /* If the filename didn't change, do nothing. */
  if (strcmp(job->info->filename, job->destinfo->filename) == 0)
    return;
  /* If the new filename does already exist, do nothing. */
  if (stat(job->destinfo->filename, &trash) == 0) {
    strncpy(job->info->status, _("File exists."), 511);
    return;
  }
  if (rename(job->info->filename, job->destinfo->filename) != 0) {
    strncpy(job->info->status, _("Rename failed."), 511);
    return;
  }
  strcpy(job->info->filename, job->destinfo->filename);
  strncpy(job->info->status, _("Done."), 511);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


void job_append(FileInfo *info, JobArg *arg)
{
  Job *job = job_new(info);
  
  ruleset_ref(arg->ruleset);
  job->ruleset = arg->ruleset;
  job->preview = arg->preview;
  
  pthread_mutex_lock(&engine.mut);
  engine.jobs = g_list_append(engine.jobs, job);
  pthread_mutex_unlock(&engine.mut);
}


/*
 * Proceed one job.
 */
void job_proceed(Job *job)
{
  Tag emptytag;
  
//  printf("INFO-PRCDEST-PRE: %s\n", job->destinfo->filename);
  /* Apply the ruleset to the current item. */
  if (!job->generated) {
    memset(job->ruleset->buffer, 0, 1024);
    g_list_foreach(job->ruleset->statements, (GFunc)statement_proceed, job);
    job->generated = TRUE;
    job->updated = TRUE;
  }
  
  /* If we are only to create a preview, do so. */
  if (job->preview) {
    /* Append the preview to a list. */
    pthread_mutex_lock(&engine.mut);
    engine.previewlist = g_list_append(engine.previewlist, job);
    pthread_mutex_unlock(&engine.mut);
    return;
  }
  
  /* Rename the file, if the new filename does not yet exist. */
  file_rename(job);
  
  memset(&emptytag, 0, sizeof(Tag));
  
  switch (job->destinfo->tag_type) {
  case ID3:
    /* If there hasn't been a tag before and there's a non-empty tag now, save it. */
    /* If there has been a tag before and it has changed, save it. */
    if ((!job->info->tag[0] && memcmp(&emptytag, job->destinfo->tag[0], sizeof(Tag)) != 0)
      || (job->info->tag[0] && memcmp(job->info->tag[0], job->destinfo->tag[0], sizeof(Tag)) != 0))
      set_id3v1_tag(job->destinfo->tag[0], job->info->filename);
    if ((!job->info->tag[1] && memcmp(&emptytag, job->destinfo->tag[1], sizeof(Tag)) != 0)
      || (job->info->tag[1] && memcmp(job->info->tag[1], job->destinfo->tag[1], sizeof(Tag)) != 0))
      set_id3v2_tag(job->destinfo->tag[1], job->info->filename, cfg.use_unicode);
    break;
    
#ifdef ENABLE_OGG
  case OGG:
    if ((!job->info->tag[0] && memcmp(&emptytag, job->destinfo->tag[0], sizeof(Tag)) != 0)
      || (job->info->tag[0] && memcmp(job->info->tag[0], job->destinfo->tag[0], sizeof(Tag)) != 0))
      set_vorbis_tag(job->destinfo->tag[0], job->info->filename);
    break;
#endif
  }
  
  /* Update the fileinfo. */
  fileinfo_read_1(job->info->filename, job->info);
  job_destroy(job);
}


/*
 * Proceed a job but don't apply it to the file. Instead,
 * append the updated fileinfo to a list.
 */
void job_preview(Job *job)
{
  /* Apply the ruleset. */
  g_list_foreach(job->ruleset->statements, (GFunc)statement_proceed, job);
  /* Prevent the preview from being destroyed. */
  job->destinfo = NULL;
  /* Destroy the job. */
  job_destroy(job);
}


void job_proceed_all(void)
{
  GList *list = NULL;
  
  if (!engine.jobs)
    return;
  
  list = engine.jobs;
  engine.jobs = NULL;
  
  pthread_mutex_unlock(&engine.mut);
  g_list_foreach(list, (GFunc)job_proceed, NULL);
  g_list_free(list);
  pthread_mutex_lock(&engine.mut);
}


/*
 * Destroy all jobs that are in the "preview" list and free the list.
 */
void job_preview_clear(void)
{
  GList *list = NULL;
  
  pthread_mutex_lock(&engine.mut);
  if (!engine.previewlist) {
    pthread_mutex_unlock(&engine.mut);
    return;
  }
  
  list = engine.previewlist;
  engine.previewlist = NULL;
  pthread_mutex_unlock(&engine.mut);
  
  if (list) {
    g_list_foreach(list, (GFunc)job_destroy, NULL);
    g_list_free(list);
  }
}


void job_remove_from_preview(Job *job)
{
  g_return_if_fail(job != NULL);
  pthread_mutex_lock(&engine.mut);
  engine.previewlist = g_list_remove(engine.previewlist, job);
  pthread_mutex_unlock(&engine.mut);
  job_destroy(job);
}


void job_previewlist_execute(void)
{
  GList *list = NULL;
  
  pthread_mutex_lock(&engine.mut);
  list = engine.previewlist;
  engine.previewlist = NULL;
  pthread_mutex_unlock(&engine.mut);
  
  g_list_foreach(list, (GFunc)job_preview_make_nonpreview, NULL);
  
  pthread_mutex_lock(&engine.mut);
  engine.jobs = g_list_concat(engine.jobs, list);
  /* Signalize that something has to be done now. */
  pthread_cond_signal(&engine.cond);
  pthread_mutex_unlock(&engine.mut);
}
