/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <libintl.h>
#include <regex.h>
#include "engine_fileinfo.h"
#include "engine_rules.h"
#include "engine_ruleset.h"
#include "lib_strings.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

// #define _DEBUG_

/***************************************************************************************
 * STATICS
 ***************************************************************************************/
/*
 * Determine the correct order for the queue columns.
 */
static short int is_spaceonly(char *name)
{
  if (*name == '\0')
    return(FALSE);

  while (*name == ' ')
    ++name;

  if (*name == '\0')
    return(TRUE);

  return(FALSE);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/***************************************************************************************
 * STATIC RULES
 ***************************************************************************************/
/*
 * This rule will replace all multiple spaces by one space.
 * Spaces are:
 * Single space, tab, vertical tab, form feed, carriage return, or newline.
 */
static RuleFunc rule_strip_multiple_spaces(char *str)
{
	const char *pstr = str;

	if (!str)
	  return;
  
  for (; *pstr != '\0'; pstr++) {
    if (!isspace(*pstr) || (isspace(*pstr) && !isspace(*(pstr + 1))))
      *str++ = *pstr;
  }
  *str = '\0';
}


/*
 * This will kill spaces at the beginning of a string.
 */
static RuleFunc rule_remove_leading_spaces(char *str)
{
#ifdef _DEBUG_
  printf("rule_remove_leading_spaces(): Before: %s\n", str);
#endif
  string_wipe_spaces(str, 0);
#ifdef _DEBUG_
  printf("rule_remove_leading_spaces(): After: %s\n", str);
#endif
}


/*
 * This will kill spaces at the end of a string.
 */
static RuleFunc rule_remove_trailing_spaces(char *str)
{
#ifdef _DEBUG_
  printf("rule_remove_trailing_spaces(): Before: %s\n", str);
#endif
  string_wipe_spaces(str, 1);
#ifdef _DEBUG_
  printf("rule_remove_trailing_spaces(): After: %s\n", str);
#endif
}


/*
 * This will kill everything before the first digit in a string.
 */
static RuleFunc rule_remove_before_number(char *str)
{
	char *pstr = str;
  
  /* Search for the first digit. */
	while (*pstr && isdigit(*pstr))
    pstr++;
  
  /* No digit found? Then do nothing. */
  if (!*pstr)
    return;
  
  strcpy(str, pstr);
}


/*
 * Replace "from" by "to".
 */
static RuleFunc rule_replace(char *str, const char *from, const char *to)
{
#ifdef _DEBUG_
  printf("Replace Before: %s\n", str);
#endif
  str_replace(str, from, to, 1023);
#ifdef _DEBUG_
  printf("Replace After : %s\n", str);
#endif
}


/*
 * Replace "from" by "to".
 */
static RuleFunc rule_replace_ci(char *str, const char *from, const char *to)
{
  str_replace_ci(str, from, to, 1023);
}


/*
 * Uppercase the whole string.
 */
static RuleFunc rule_uppercase(char *str)
{
  strtoupper(str);
}


/*
 * Lowercase the whole string.
 */
static RuleFunc rule_lowercase(char *str)
{
  strtolower(str);
}


/*
 * Uppercase the first character of every word.
 */
static RuleFunc rule_uppercase_word(char *str)
{
  strtoupper_partly(str, 1);
}


/*
 * Uppercase the first character of the first word.
 */
static RuleFunc rule_uppercase_first_word(char *str)
{
  strtoupper_partly(str, 0);
}


/*
 * Grab a substring from a field and copy it into a buffer.
 */
static RuleFunc rule_buffer_grab(char *src, char *format, char *buffer)
{
  char *match = NULL;
  
  if (shellpattern_match_grab(src, format, FALSE, &match) != 0)
    return;
  
  strncpy(buffer, match, 1023);
}


/*
 * Copy the buffer contents into a field.
 */
static RuleFunc rule_buffer_copy_to(char *src, char *trash, char *buffer)
{
  char *match = NULL;
  
  if (!buffer || *buffer == '\0')
    return;
  
  // FIXME! Buffer Overflow!
  strcpy(src, buffer);
}


/*
 * Grab several substrings from a field using a format like "%s - %a - %s" and
 * copy them into their destinations.
 *
 * %a = Artist
 * %b = Album
 * %c = Comment
 * %g = Genre
 * %y = Year
 * %s = Song
 * %t = Track
 * %x = Trash
 */
RuleFunc rule_format_to_destination(char *src, const char *format, FileInfo *info)
{
  const char  *pformat    = format;
  char        *regexp     = NULL;
  char        *pregexp    = NULL;
  regex_t     preg;               /* Pre-compiled pattern buffer. */
  int         numoffsets  = 50;
  regmatch_t  pmatch[numoffsets]; /* Substring addresses by regexec. */
  int         i           = 1;
  int         flags       = 0;
  int         len         = 0;
  
  if (!src || *src == '\0' || !format || *format == '\0')
    return;
  
  /* Reserve enough space for the regexp string. */
  regexp  = calloc(1, 2048);
  pregexp = regexp;
  
  /* Walk through the format-string, creating a parsable string */
  /* for the regexp handler. */
  while (*pformat != '\0') {
    /* Braces must be escaped. */
    if (*pformat == '(' || *pformat == ')') {
      //*(pregexp++) = '\\';
      *(pregexp++) = *(pformat++);
      continue;
    }
    
    /* Not a switch? Then just copy the current char. */
    if (*pformat != '%') {
      *(pregexp++) = *(pformat++);
      continue;
    }
    
    /* Ok, so it is a switch. Our action depends on the following char. */
    pformat++;
    switch (*pformat) {
    case '%':
      /* "%%" should only be replaced by a single "%". */
      *(pregexp++) = *pformat;
      break;
    
    case 'x':
      /* "%x" has no meaning. Drop this part to trash (regexp ".*"). */
      strcpy(pregexp, ".*");
      pregexp += 2;
      break;
    
    case 't':
    case 'T':
      /* The Track matches digits only. */
      strcpy(pregexp, "\\([0-9]*\\)");
      pregexp += 10;
      break;
      
    case 'a':
    case 'b':
    case 'c':
    case 'g':
    case 'y':
    case 's':
      /* One-Word match. */
      strcpy(pregexp, "\\([^ ]*\\)");
      pregexp += 9;
      break;
    
    case 'A':
    case 'B':
    case 'C':
    case 'G':
    case 'Y':
    case 'S':
      /* Greedy match. */
      strcpy(pregexp, "\\(.*\\)");
      pregexp += 6;
      break;
    
    default:
      /* Other switches are not allowed. */
      free(regexp);
      return;
      break;
    }
    
    pformat++;
  }
  
  /* Ok, now we have the regexp string. */
  /* Compile the regular expression. */
  flags = REG_NEWLINE;
  if (regcomp(&preg, regexp, flags) != 0) {
    fprintf(stderr, "Error: RegExp compilation failed: \"%s\"\n", regexp);
    free(regexp);
    return;
  }
  
  //printf("Regexp: %s\n", regexp);
  
  free(regexp);
  
  /* Execute the regular expression. */
  if (regexec(&preg, src, numoffsets, pmatch, REG_NOTBOL | REG_NOTEOL) != 0)
    return;
  
  /* Walk through the format-string again, copying all substrings */
  /* into the destination fields. */
  pformat = format;
  while (*pformat != '\0') {
    /* Not a switch? Then just move on. */
    if (*pformat != '%') {
      pformat++;
      continue;
    }
    
    /* Ok, so it is a switch. Our action depends on the following char. */
    pformat++;
    
    if (*pformat == '%' || *pformat == 'x' || *pformat == 'X') {
      /* "%%" and "%x" are being ignored by us. */
      pformat++;
      continue;
    }
    
    /* Ok, so it is a *valid* switch. Calculate the substring length. */
    len = (pmatch[i].rm_eo - pmatch[i].rm_so);
    
    /* Make sure that the regexp handler has fetched that substring. */
    if (len <= 0) {
      i++;
      pformat++;
      continue;
    }
    
    switch (tolower(*pformat)) {
    case 'a':
      if (len >= 1024)
        len = 1023;
      memcpy(info->tag[0]->artist, src + pmatch[i].rm_so, len);
      *(info->tag[0]->artist + len) = '\0';
      strcpy(info->tag[1]->artist, info->tag[0]->artist);
      break;
      
    case 'b':
      if (len >= 1024)
        len = 1023;
      memcpy(info->tag[0]->album, src + pmatch[i].rm_so, len);
      *(info->tag[0]->album + len) = '\0';
      strcpy(info->tag[1]->album, info->tag[0]->album);
      break;
      
    case 'c':
      if (len >= 1024)
        len = 1023;
      memcpy(info->tag[0]->comment, src + pmatch[i].rm_so, len);
      *(info->tag[0]->comment + len) = '\0';
      strcpy(info->tag[1]->comment, info->tag[0]->comment);
      break;
      
    case 'g':
      if (len >= 512)
        len = 511;
      memcpy(info->tag[0]->genre, src + pmatch[i].rm_so, len);
      *(info->tag[0]->genre + len) = '\0';
      strcpy(info->tag[1]->genre, info->tag[0]->genre);
      break;
      
    case 'y':
      if (len >= 5)
        len = 4;
      memcpy(info->tag[0]->year, src + pmatch[i].rm_so, len);
      *(info->tag[0]->year + len) = '\0';
      strcpy(info->tag[1]->year, info->tag[0]->year);
      break;
      
    case 's':
      if (len >= 1024)
        len = 1023;
      memcpy(info->tag[0]->title, src + pmatch[i].rm_so, len);
      *(info->tag[0]->title + len) = '\0';
      strcpy(info->tag[1]->title, info->tag[0]->title);
      break;
      
    case 't':
      if (len >= 20)
        len = 19;
      memcpy(info->tag[0]->track, src + pmatch[i].rm_so, len);
      *(info->tag[0]->track + len) = '\0';
      strcpy(info->tag[1]->track, info->tag[0]->track);
      break;
      
    default:
      return;
      break;
    }
    
    i++;
    pformat++;
  }
}
/***************************************************************************************
 * END STATIC RULES
 ***************************************************************************************/


static void *rulefunc_data_get(int row, int col)
{
  typedef struct {
    int       type;
    char      *description;
    RuleFunc  func;
  } RuleFuncStruct;
  RuleFuncStruct rulelist[] = {
    /* Flags. */
    { RULE_STRIP_MULTIPLE_SPACES,   N_("Strip multiple spaces."),                           (RuleFunc)rule_strip_multiple_spaces },
    { RULE_REMOVE_LEADING_SPACES,   N_("Strip leading spaces."),                            (RuleFunc)rule_remove_leading_spaces },
    { RULE_REMOVE_TRAILING_SPACES,  N_("Strip trailing spaces."),                           (RuleFunc)rule_remove_trailing_spaces },
    { RULE_REMOVE_BEFORE_NUMBER,    N_("Remove every character before the first number."),  (RuleFunc)rule_remove_before_number },
    
    /* Capitol Letters. */
    { RULE_UPPERCASE,               N_("Uppercase everything."),                            (RuleFunc)rule_uppercase },
    { RULE_UPPERCASE_WORD,          N_("Uppercase the first character of every word."),     (RuleFunc)rule_uppercase_word },
    { RULE_UPPERCASE_FIRST_WORD,    N_("Uppercase the first character of the first word."), (RuleFunc)rule_uppercase_first_word },
    { RULE_LOWERCASE,               N_("Lowercase everything."),                            (RuleFunc)rule_lowercase },
    
    /* Static. */
    { RULE_SET,                     N_("Set value \"%s\"."),                                (RuleFunc)strcpy },
    
    /* Modifiers. */
    { RULE_REPLACE,                 N_("Replace (case sensitive) \"%s\" by \"%s\"."),       (RuleFunc)rule_replace },
    { RULE_REPLACE_CI,              N_("Replace (case insensitive) \"%s\" by \"%s\"."),     (RuleFunc)rule_replace_ci },
    
    /* Buffer Operations. */
    { RULE_BUFFER_GRAB,             N_("Grab a string into the buffer using \"%s\"."),      (RuleFunc)rule_buffer_grab },
    { RULE_BUFFER_COPY_TO,          N_("Copy the buffer into the field."),                  (RuleFunc)rule_buffer_copy_to },
    
    { RULE_FORMAT_GRAB,             N_("Generate a tag using the format string \"%s\"."),   (RuleFunc)rule_format_to_destination },
    
    { -1, NULL, NULL }
  };
  switch (col) {
  case 0:
    return((void*)rulelist[row].type);
  
  case 1:
    return(_(rulelist[row].description));
  
  case 2:
    return(rulelist[row].func);
  }
}


void rule_describe(Rule *rule, void *arg1, void *arg2)
{
  int i = -1;
  
  g_assert(rule != NULL);
  
  while (rulefunc_data_get(++i, 1)) {
    if ((int)rulefunc_data_get(i, 0) != rule->type)
      continue;
    snprintf(rule->descr, 1023, (char*)rulefunc_data_get(i, 1), (char *)arg1, (char *)arg2);
    break;
  }
}


/*
 * Apply a rule that changes one field only.
 */
void rule_apply(int type, FileInfo *info, int field, void *data1, void *data2)
{
  int tag = (field & (T_ID3V1 | T_ID3V2 | T_OGG)) >> 11;
  typedef struct {
    int       field;
    char      *dest;
  } Pair;
  Pair map[] = {
    { F_FILENAME, strrchr(info->filename, '/') ? strrchr(info->filename, '/') + 1 : info->filename },
    { F_ARTIST,   info->tag[tag]->artist },
    { F_SONG,     info->tag[tag]->title },
    { F_ALBUM,    info->tag[tag]->album },
    { F_TRACK,    info->tag[tag]->track },
    { F_YEAR,     info->tag[tag]->year },
    { F_GENRE,    info->tag[tag]->genre },
    { F_COMMENT,  info->tag[tag]->comment },
    { 0, NULL }
  };
  int i = -1;
  RuleFunc func;
  
  /* Walk through the function list, looking for the matching function. */
  while (rulefunc_data_get(++i, 1)) {
    if ((int)rulefunc_data_get(i, 0) == -1) {
      fprintf(stderr, "Error: rule_apply(): Invalid rule identifier found!\n");
      return;
    }
    if ((int)rulefunc_data_get(i, 0) != type)
      continue;
    func = (RuleFunc)rulefunc_data_get(i, 2);
  }
  
  /* Walk through the field list, applying the rule to every matching field. */
  i = -1;
  while (map[++i].dest) {
    if ((field & map[i].field) != map[i].field)
      continue;
    func(map[i].dest, data1, data2);
  }
}
