/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include "engine_fileinfo.h"
#include "engine_conditions.h"
#include "engine_rules.h"
#include "engine_jobs.h"
#include "engine_ruleset.h"
#include "engine_init.h"
#include "lib_mpgheader.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern Cfg cfg;

typedef struct {
  short int dest;
  char      *name;
} Pair;

static Pair fielddescr[] = {
  { F_FILENAME,  N_("Filename") },
  { F_ARTIST,    N_("Artist") },
  { F_SONG,      N_("Song") },
  { F_ALBUM,     N_("Album") },
  { F_TRACK,     N_("Track") },
  { F_YEAR,      N_("Year") },
  { F_GENRE,     N_("Genre") },
  { F_COMMENT,   N_("Comment") },
  { 0,           NULL }
};


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static RuleStatement *statement_new(void)
{
  return((RuleStatement *)calloc(1, sizeof(RuleStatement)));
}


static RuleStatement *statement_copy(RuleStatement *sm)
{
  RuleStatement *smcp = statement_new();
  
  g_assert(sm != NULL);
  memcpy(smcp, sm, sizeof(RuleStatement));
  (sm->cond.data1type == INT16) ? (smcp->cond.data1 = sm->cond.data1) : (smcp->cond.data1 = strdup(sm->cond.data1));
  (sm->rule.data1type == INT16) ? (smcp->rule.data1 = sm->rule.data1) : (smcp->rule.data1 = strdup(sm->rule.data1));
  (sm->rule.data2type == INT16) ? (smcp->rule.data2 = sm->rule.data2) : (smcp->rule.data2 = strdup(sm->rule.data2));
  return(smcp);
}


static void statement_destination_describe(RuleStatement *statement)
{
  int i = -1;
  
  memset(statement->dest_descr, 0, 1024);
  while (fielddescr[++i].name) {
    if (!(statement->dest & fielddescr[i].dest))
      continue;
    
    if (*statement->dest_descr == '\0')
      strncpy(statement->dest_descr, _(fielddescr[i].name), 1023);
    else
      snprintf(statement->dest_descr + strlen(statement->dest_descr), 1023 - strlen(statement->dest_descr),
        ", %s", _(fielddescr[i].name));
  }
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Create a new ruleset.
 */
Ruleset *ruleset_new(const char *name)
{
  Ruleset *ruleset = (Ruleset *)calloc(1, sizeof(Ruleset));
  if (!name)
    return(ruleset);
  strncpy(ruleset->name, name, 1023);
  ruleset_ref(ruleset);
  return(ruleset);
}


/*
 * Create a reference to a ruleset.
 */
void ruleset_ref(Ruleset *ruleset)
{
  ruleset->ref++;
}


/*
 * Remove a reference from a ruleset and destroy it if the refcounter is < 0.
 */
void ruleset_unref(Ruleset *ruleset)
{
  if (--(ruleset->ref) > 0)
    return;
  
  g_list_foreach(ruleset->statements, (GFunc)statement_destroy, NULL);
  g_list_free(ruleset->statements);
  free(ruleset);
}


void ruleset_clear(Ruleset *ruleset)
{
  g_list_foreach(ruleset->statements, (GFunc)statement_destroy, NULL);
  g_list_free(ruleset->statements);
  ruleset->statements = NULL;
}


Ruleset *ruleset_copy(Ruleset *ruleset)
{
  Ruleset *rulesetcp  = (Ruleset *)calloc(1, sizeof(Ruleset));
  GList   *item       = NULL;
  
  g_assert(ruleset != NULL);
  
  strncpy(rulesetcp->name, ruleset->name, 1024);
  ruleset_ref(rulesetcp);
  
  item = ruleset->statements;
  while (item) {
    if (!item->data) {
      item = item->next;
      continue;
    }
    rulesetcp->statements = g_list_append(rulesetcp->statements, statement_copy((RuleStatement*)item->data));
    item = item->next;
  }
  
  return(rulesetcp);
}


/*
 * Find a ruleset with the given name and return a pointer to it.
 * Returns NULL if no such list exists.
 */
Ruleset *ruleset_find(GList *rulelists, const char *name)
{
  while (rulelists) {
    if (!rulelists->data || !((Ruleset *)rulelists->data)->name) {
      rulelists = rulelists->next;
      continue;
    }
    if (strcmp(((Ruleset *)rulelists->data)->name, name) == 0)
      return((Ruleset *)rulelists->data);
    rulelists = rulelists->next;
  }
  return(NULL);
}


/*
 * Apply one statement.
 */
GFunc statement_proceed(RuleStatement *sm, void *data)
{
  Job *job = (Job *)data;
  
  /* Execute the condition. */
  if (!cond_is_true(sm->cond.type, job->destinfo, sm->cond.field & F_FILENAME, sm->cond.data1))
    return;
  if (!cfg.defaulttag_v2) {
    if (!cond_is_true(sm->cond.type, job->destinfo, T_ID3V1 | (sm->cond.field & (FIELDMASK ^ F_FILENAME)), sm->cond.data1))
      return;
  }
  else {
    if (!cond_is_true(sm->cond.type, job->destinfo, T_ID3V2 | (sm->cond.field & (FIELDMASK ^ F_FILENAME)), sm->cond.data1))
      return;
  }
  
  /* Execute the rule. */
  switch (sm->rule.type) {
  case RULE_BUFFER_GRAB:
  case RULE_BUFFER_COPY_TO:
    /* This rule requires special handling as the destination is the ruleset buffer. */
    rule_apply(sm->rule.type, job->destinfo, sm->dest & F_FILENAME, sm->rule.data1, job->ruleset->buffer);
    if (!cfg.defaulttag_v2)
      rule_apply(sm->rule.type, job->destinfo, T_ID3V1 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, job->ruleset->buffer);
    else
      rule_apply(sm->rule.type, job->destinfo, T_ID3V2 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, job->ruleset->buffer);
    break;
  
  case RULE_FORMAT_GRAB:
    /* This rule requires special handling as the destination is the fileinfo struct. */
    rule_apply(sm->rule.type, job->destinfo, sm->dest & F_FILENAME, sm->rule.data1, job->destinfo);
    if (!cfg.defaulttag_v2)
      rule_apply(sm->rule.type, job->destinfo, T_ID3V1 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, job->destinfo);
    else
      rule_apply(sm->rule.type, job->destinfo, T_ID3V2 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, job->destinfo);
    break;
    
  default:
    /* Apply to the filename. */
    rule_apply(sm->rule.type, job->destinfo, sm->dest & F_FILENAME, sm->rule.data1, sm->rule.data2);
    /* Apply to other fields. */
    rule_apply(sm->rule.type, job->destinfo, T_ID3V1 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, sm->rule.data2);
    rule_apply(sm->rule.type, job->destinfo, T_ID3V2 | (sm->dest & (FIELDMASK ^ F_FILENAME)), sm->rule.data1, sm->rule.data2);
  }
  
  return(0);
}


/*
 * Append a new rule to a ruleset.
 */
void statement_append(Ruleset *ruleset,
  int dest,
  int condition,
  int cond_field,
  short int conddata1type, void *conddata1,
  int rule,
  short int ruledata1type, void *ruledata1,
  short int ruledata2type, void *ruledata2)
{
  RuleStatement *statement = statement_new();
  
  /* Create the condition. */
  statement->cond.type = condition;
  statement->cond.field = cond_field;
  statement->cond.data1type = conddata1type;
  
  if (conddata1type == INT16)
    statement->cond.data1 = conddata1;
  else if (conddata1type != INT16 && conddata1)
    statement->cond.data1 = strdup(conddata1);
  else if (conddata1type != INT16 && conddata1)
    statement->cond.data1 = NULL;
  
  condition_describe(&statement->cond, conddata1);
  
  /* Create the rule. */
  statement->rule.type = rule;
  statement->rule.data1type = ruledata1type;
  statement->rule.data2type = ruledata2type;
  
  if (ruledata1type == INT16)
    statement->rule.data1 = ruledata1;
  else if (ruledata1type != INT16 && ruledata1)
    statement->rule.data1 = strdup(ruledata1);
  else if (ruledata1type != INT16 && ruledata1)
    statement->rule.data1 = NULL;
  
  if (ruledata2type == INT16)
    statement->rule.data2 = ruledata2;
  else if (ruledata2type != INT16 && ruledata2)
    statement->rule.data2 = strdup(ruledata2);
  else if (ruledata2type != INT16 && ruledata2)
    statement->rule.data2 = NULL;
  
  rule_describe(&statement->rule, ruledata1, ruledata2);
  
  /* Create the destination description. */
  statement->dest = dest;
  statement_destination_describe(statement);
  
  ruleset->statements = g_list_append(ruleset->statements, statement);
}


void statement_remove(Ruleset *ruleset, RuleStatement *statement)
{
  ruleset->statements = g_list_remove(ruleset->statements, statement);
  statement_destroy(statement);
}


void statement_destroy(RuleStatement *statement)
{
  if (statement->cond.data1type == STR)
    free(statement->cond.data1);
  if (statement->rule.data1type == STR)
    free(statement->rule.data1);
  if (statement->rule.data2type == STR)
    free(statement->rule.data2);
  free(statement);
}


/*
 * Save a whole list of rulesets to a configfile.
 */
short int rulesetlist_save(GList *rulesetlist)
{
  Ruleset       *ruleset    = NULL;
  RuleStatement *statement  = NULL;
  GList         *item       = NULL;
  int           i           = 1;
  int           f           = 1;
  char          config[50000];
  char          filename[1024];
  char          curcat[1024];
  char          curoption[1024];
  char          curvalue[1024];
  char          curstring[1024];
  
  memset(config, 0, sizeof(config));
  snprintf(filename, 1023, "%s/" RULESETLISTFILE, getenv("HOME"));
  
  /* Walk through all rulesets. */
  while (rulesetlist) {
    ruleset = (Ruleset*)rulesetlist->data;
    
    snprintf(curcat, 1023, "Ruleset %i", i++);
    configfile_option_add(config, curcat, "rulesetname", ruleset->name);
    
    /* Walk through the current ruleset's statements. */
    item = ruleset->statements;
    f = 1;
    while (item) {
      statement = (RuleStatement*)item->data;
      
      /* Save the condition. */
      snprintf(curstring, 1023, "%%i\t%%i\t%%i\t%%%c",
        (statement->cond.data1type == INT16) ? 'i' : 's');
      snprintf(curoption, 1023, "statement %i cond", f);
      snprintf(curvalue, 1023, curstring, statement->cond.type, statement->cond.field,
        statement->cond.data1type, statement->cond.data1);
      str_replace(curvalue, "[s]", "[_s_]", 1023);
      str_replace(curvalue, " ", "[s]", 1023);
      configfile_option_add(config, curcat, curoption, curvalue);
      
      /* Save the rule. */
      snprintf(curstring, 1023, "%%i\t%%i\t%%%c\t%%i\t%%%c",
        (statement->rule.data1type == INT16) ? 'i' : 's',
        (statement->rule.data2type == INT16) ? 'i' : 's');
      snprintf(curoption, 1023, "statement %i rule", f);
      snprintf(curvalue, 1023, curstring, statement->rule.type,
      statement->rule.data1type, statement->rule.data1,
      statement->rule.data2type, statement->rule.data2);
      str_replace(curvalue, "[s]", "[_s_]", 1023);
      str_replace(curvalue, " ", "[s]", 1023);
      configfile_option_add(config, curcat, curoption, curvalue);
      
      /* Save the destination. */
      snprintf(curoption, 1023, "statement %i dest", f);
      snprintf(curvalue, 1023, "%i", statement->dest);
      configfile_option_add(config, curcat, curoption, curvalue);
      
      item = item->next;
      f++;
    }
    
    rulesetlist = rulesetlist->next;
  }
  
  return(configfile_write(filename, config));
}


/*
 * Load a list of rulesets from a configfile into memory.
 */
short int rulesetlist_load(GList **rulesetlist)
{
  Ruleset       *ruleset    = NULL;
  GList         *item       = NULL;
  int           i           = 1;
  int           f           = 1;
  char          config[50000];
  char          filename[1024];
  char          curcat[1024];
  char          curoption[1024];
  char          curcondvalue[2048];
  char          currulevalue[2048];
  char          curdestvalue[2048];
  int           ruletype, dest, condtype, condfield;
  int           ruledata1type, ruledata2type;
  char          ruledata1[1024], ruledata2[1024];
  int           conddata1type;
  char          conddata1[1024];
  
  /* Read the configfile. */
  memset(config, 0, sizeof(config));
  snprintf(filename, 1023, "%s/" RULESETLISTFILE, getenv("HOME"));
  configfile_read(filename, config, 50000);
  
  /* Read all rulesets. */
  while (1) {
    snprintf(curcat, 1023, "Ruleset %i", i);
    if (!configfile_option_get(config, curcat, "rulesetname", currulevalue))
      break;
    
    ruleset = ruleset_new(currulevalue);
    
    f = 1;
    while (1) {
      /* Read the condition. */
      snprintf(curoption, 1023, "statement %i cond", f);
      if (!configfile_option_get(config, curcat, curoption, curcondvalue))
        break;
      
      /* Read the rule. */
      snprintf(curoption, 1023, "statement %i rule", f);
      if (!configfile_option_get(config, curcat, curoption, currulevalue))
        break;
      
      /* Read the destination. */
      snprintf(curoption, 1023, "statement %i dest", f);
      if (!configfile_option_get(config, curcat, curoption, curdestvalue))
        break;
      
      /* Put the read data into a new statement. */
      sscanf(curcondvalue, "%d\t%d\t%d\t%s", &condtype, &condfield, &conddata1type, &conddata1);
      sscanf(currulevalue, "%d\t%d\t%s\t%d\t%s", &ruletype, &ruledata1type, &ruledata1, &ruledata2type, &ruledata2);
      sscanf(curdestvalue, "%d", &dest);
      
      str_replace(conddata1, "[s]", " ", 1023);
      str_replace(conddata1, "[_s_]", "[s]", 1023);
      str_replace(ruledata1, "[s]", " ", 1023);
      str_replace(ruledata1, "[_s_]", "[s]", 1023);
      str_replace(ruledata2, "[s]", " ", 1023);
      str_replace(ruledata2, "[_s_]", "[s]", 1023);
      
      statement_append(ruleset, dest,
        condtype,
        condfield,
        conddata1type, (conddata1type == 1) ? (void *)atoi(conddata1) : conddata1,
        ruletype,
        ruledata1type, (ruledata1type == 1) ? (void *)atoi(ruledata1) : ruledata1,
        ruledata2type, (ruledata2type == 1) ? (void *)atoi(ruledata2) : ruledata2);
      
      f++;
    }
    
    /* If we were unsuccessful, destroy the ruleset. */
    if (f <= 1) {
      ruleset_unref(ruleset);
      i++;
      continue;
    }
    
    *rulesetlist = g_list_append(*rulesetlist, ruleset);
    
    i++;
  }
  
  if (i <= 1)
    return(FALSE);
  
  return(TRUE);
}


/*
 * Return the name of the given field id (Artist, Filename, Song, ...)
 */
const char *field_name_get(int type)
{
  int i = -1;
  
  while (fielddescr[++i].name) {
    if (fielddescr[i].dest == type)
      return(fielddescr[i].name);
  }
  
  return(NULL);
}
