/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <libintl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include "engine_init.h"
#include "engine_queue.h"
#include "gui_init.h"
#include "gui_queue.h"
#include "gui_addfiles.h"
#include "gui_lib_dirtree.h"
#include "gui_lib_filelist.h"
#include "gui_lib_popup.h"
#include "lib_recurse.h"
#include "lib_shellpattern.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern Engine engine;
extern GUI gui;
extern Cfg cfg;

typedef struct {
  const gchar *filepattern;
  short int   filecase;
  const gchar *dirpattern;
  short int   dircase;
  short int   needv1;
  short int   neednov1;
  short int   needv2;
  short int   neednov2;
  short int   needogg;
  short int   neednoogg;
} CondPattern;


enum TYPES {
  TOGGLE,
  ENTRY,
  OPTIONMENU
};

typedef struct {
  gchar *widgetname;
  gint  type;
  void  *value;
} Map;

static Map options[] = {
  { "check_addfiles_directories_recursive",                     TOGGLE, &gui.addfiles.recursive },
  
  /* Filename filters. */
  { "check_addfiles_directories_filename_case",                 TOGGLE, &gui.addfiles.filename_casesensitive },
  { "entry_addfiles_directories_filename_pattern",              ENTRY,  &gui.addfiles.filename_pattern },
  
  { "check_addfiles_directories_directory_case",                TOGGLE, &gui.addfiles.dirname_casesensitive },
  { "entry_addfiles_directories_directory_pattern",             ENTRY,  &gui.addfiles.dirname_pattern },
  
  /* Tag filters. */
  { "check_addfiles_tagfilter_id3v1",                           TOGGLE, &gui.addfiles.tagfilter_id3v1 },
  { "radio_addfiles_tagfilter_id3v1_required",                  TOGGLE, &gui.addfiles.tagfilter_id3v1_req },
  
  { "check_addfiles_tagfilter_id3v2",                           TOGGLE, &gui.addfiles.tagfilter_id3v2 },
  { "radio_addfiles_tagfilter_id3v2_required",                  TOGGLE, &gui.addfiles.tagfilter_id3v2_req },
  
  { "check_addfiles_tagfilter_ogg",                             TOGGLE, &gui.addfiles.tagfilter_ogg },
  { "radio_addfiles_tagfilter_ogg_required",                    TOGGLE, &gui.addfiles.tagfilter_ogg_req },
  
  { NULL, 0, NULL }
};


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static void add_to_queue_if_pattern_matches(const gchar *fullfilename, void *ppattern)
{
  CondPattern *pattern = ppattern;
  gchar       *filename = NULL;
  gchar       *dir = NULL;
  Tag         dummytag;
  short int   havev1  = FALSE;
  short int   havev2  = FALSE;
  short int   haveogg = FALSE;
  
  filename = strdup(strrchr(fullfilename, '/') + 1);
  dir = strdup(fullfilename);
  *strrchr(dir, '/') = '\0';
  
  /* Check if the filename filter matches. */
  if (!shellpattern_match(filename, pattern->filepattern, pattern->filecase)
    || !shellpattern_match(dir, pattern->dirpattern, pattern->dircase))
    goto done;
  
  /* Check if the tag filters must be applied (and apply them). */
  if (pattern->needv1 || pattern->neednov1
    || pattern->needv2 || pattern->neednov2
    || pattern->needogg || pattern->neednoogg) {
    /* Determine which tags exist. */
    havev1 = (get_id3v1_tag(&dummytag, fullfilename) == 0);
    havev2 = (get_id3v2_tag(&dummytag, fullfilename) == 0);
#ifdef ENABLE_OGG
    haveogg = (get_vorbis_tag(&dummytag, fullfilename) == 0);
#endif
    
    /* Determine if that matches the user's wishes. */
    if (pattern->needv1 && !havev1)
      goto done;
    if (pattern->neednov1 && havev1)
      goto done;
    if (pattern->needv2 && !havev2)
      goto done;
    if (pattern->neednov2 && havev2)
      goto done;
    if (pattern->needogg && !haveogg)
      goto done;
    if (pattern->neednoogg && haveogg)
      goto done;
  }
  
  /* Ending up here, we want that file ;). */
  queue_append(fullfilename);
  
  /* Cleanups and GUI update. */
done:
  gdk_threads_leave();
  while(g_main_iteration(FALSE));
  gdk_threads_enter();
  
  g_free(filename);
  g_free(dir);
}


/*
 * Returns TRUE if *path is a directory, FALSE if not.
 */
static short int is_dir(char *path)
{
  struct stat pstat;
  
  if (stat(path, &pstat) < 0)
    return(FALSE);
  
  return(S_ISDIR(pstat.st_mode) != 0);
}


/*
 * Call a function for each file in "dir" that matches one of the patterns in "pattern[]".
 */
static void directory_for_each_file(const char *directory, const char *pattern[], void (* func)(const char *filename, void *data), void *data)
{
  DIR *stream = NULL;
  struct dirent *filestruct = NULL;
  char fullfilename[4096];
  
  /* Open the directory. */
  if (!(stream = opendir(directory))) {
    printf("Error: recursive_for_each_file(): Failed opening path. (%s)\n", directory);
    return;
  }
  
  /* Parse through the directory. */
  while ((filestruct = readdir(stream))) {
    /* Create a fullpath. */
    snprintf(fullfilename, 4095, "%s%s", directory, filestruct->d_name);
    
    /* Ignore directories. */
    if (is_dir(fullfilename))
      continue;
    
    /* Does the file match the pattern? */
    if (pattern && !shellpattern_match_any(filestruct->d_name, pattern, FALSE))
      continue;
    
    /* Call the user's function. */
    func(fullfilename, data);
  }
  
  closedir(stream);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/***************************************************************************************
 * CALLBACKS
 ***************************************************************************************/
void on_treeview_addfiles_files_directories_row_expanded(
  GtkTreeView *treeview,
  GtkTreeIter *iter,
  GtkTreePath *path,
  gpointer trash)
{
  dirtree_expand(gtk_tree_view_get_model(treeview), path, gui.showhidden);
}


void on_treeview_addfiles_files_directories_row_collapsed(
  GtkTreeView *treeview,
  GtkTreeIter *iter,
  GtkTreePath *path,
  gpointer trash)
{
  dirtree_collapse(gtk_tree_view_get_model(treeview), path, TRUE);
}


gboolean on_treeview_addfiles_files_directories_button_press_event(
  GtkWidget *tree,
  GdkEventButton *event,
  gpointer trash)
{
  GtkTreeModel *model = NULL;
  GtkTreePath *path = NULL;
  const gchar *dir = NULL;
  
  switch (event->type) {
  case GDK_2BUTTON_PRESS:
    gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(tree), event->x, event->y, &path, NULL, NULL, NULL);
    if (gtk_tree_view_row_expanded(GTK_TREE_VIEW(tree), path))
      gtk_tree_view_collapse_row(GTK_TREE_VIEW(tree), path);
    else
      gtk_tree_view_expand_row(GTK_TREE_VIEW(tree), path, FALSE);
    gtk_tree_path_free(path);
    
  case GDK_BUTTON_PRESS:
    gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(tree), event->x, event->y, &path, NULL, NULL, NULL);
    model = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(tree)));
    
    dir = dirtree_path_get(GTK_TREE_VIEW(tree), path);
    gtk_tree_path_free(path);
    
    filelist_update(gui.addfiles.files_filelist_store, dir, (const gchar**)cfg.file_pattern, gui.showhidden);
    break;
    
  default:
    break;
  }
  
  return(FALSE);
}


void on_treeview_addfiles_files_directories_row_activated(
  GtkTreeView *treeview,
  GtkTreePath *path,
  GtkTreeViewColumn *column,
  gpointer trash)
{
  const gchar *dir = NULL;
  
  dir = dirtree_path_get(GTK_TREE_VIEW(treeview), path);
  filelist_update(gui.addfiles.files_filelist_store, dir, (const gchar**)cfg.file_pattern, gui.showhidden);
}


void on_treeview_addfiles_directories_directories_row_collapsed(
  GtkTreeView *treeview,
  GtkTreeIter *iter,
  GtkTreePath *path,
  gpointer trash)
{
  dirtree_collapse(gtk_tree_view_get_model(treeview), path, TRUE);
}


void on_treeview_addfiles_directories_directories_row_expanded(
  GtkTreeView *treeview,
  GtkTreeIter *iter,
  GtkTreePath *path,
  gpointer trash)
{
  dirtree_expand(gtk_tree_view_get_model(treeview), path, gui.showhidden);
}


void on_button_addfiles_close_clicked(GtkButton *button, gpointer trash)
{
  gtk_widget_hide(gui.addfiles.widget);
}


void on_button_addfiles_addfiles_clicked(GtkButton *button, gpointer trash)
{
  gint currenttab = gtk_notebook_get_current_page(GTK_NOTEBOOK(glade_xml_get_widget(gui.addfiles.xml, "notebook_addfiles")));
  const gchar *dir = NULL;
  CondPattern pattern;
  
  /* If the engine is working, don't add anything. */
  if (engine.progress != -1) {
    gui_popup(gui.addfiles.widget, _("The program is still working.\nPlease wait and try again."), GTK_MESSAGE_WARNING);
    return;
  }
  
  gui.working = TRUE;
  gdk_threads_leave();
  while(g_main_iteration(FALSE));
  gdk_threads_enter();
  
  pattern.filecase    = gui.addfiles.filename_casesensitive;
  pattern.dircase     = gui.addfiles.dirname_casesensitive;
  pattern.filepattern = gui.addfiles.filename_pattern;
  pattern.dirpattern  = gui.addfiles.dirname_pattern;
  pattern.needv1      = (gui.addfiles.tagfilter_id3v1 && gui.addfiles.tagfilter_id3v1_req);
  pattern.neednov1    = (gui.addfiles.tagfilter_id3v1 && !gui.addfiles.tagfilter_id3v1_req);
  pattern.needv2      = (gui.addfiles.tagfilter_id3v2 && gui.addfiles.tagfilter_id3v2_req);
  pattern.neednov2    = (gui.addfiles.tagfilter_id3v2 && !gui.addfiles.tagfilter_id3v2_req);
  pattern.needogg     = (gui.addfiles.tagfilter_ogg   && gui.addfiles.tagfilter_ogg_req);
  pattern.neednoogg   = (gui.addfiles.tagfilter_ogg   && !gui.addfiles.tagfilter_ogg_req);
  
  /* What happens depends on which tab is active. */
  switch (currenttab) {
  case 0:
    /* Add Files. */
    filelist_selection_for_each(GTK_TREE_VIEW(glade_xml_get_widget(gui.addfiles.xml, "treeview_addfiles_files_files")), queue_append);
    break;
  
  case 1:
    if (!(dir = dirtree_path_get_selected(GTK_TREE_VIEW(glade_xml_get_widget(gui.addfiles.xml, "treeview_addfiles_directories_directories")))))
      return;
    
    /* Add recursive (opt. using filters). */
    if (gui.addfiles.recursive)
      recursive_for_each_file(dir, (const char**)cfg.file_pattern, add_to_queue_if_pattern_matches, (void *)&pattern);
    /* Add only the current dir. (opt. using filters). */
    else
      directory_for_each_file(dir, (const char**)cfg.file_pattern, add_to_queue_if_pattern_matches, (void *)&pattern);
    break;
  
  default:
    g_error("on_button_addfiles_addfiles_clicked(): Something bad happened.\n");
    break;
  }
  
  gui_queue_update(GTK_TREE_MODEL(gui.queue));
  
  gui.working = FALSE;
}


gint on_addfiles_delete_event(GtkWidget *widget, GdkEvent *event, gpointer trash)
{
  gtk_widget_hide(widget);
  return(TRUE);
}


/*
 * We use ONE function for all GUI callbacks.
 * That's an easy way to always have the updated values in a nice struct.
 */
void on_gui_addfiles_rule_widget_changed(GtkWidget *widget)
{
  const gchar *widgetname = gtk_widget_get_name(GTK_WIDGET(widget));
  gint        i = -1;
  
  g_assert(widgetname != NULL);
  
  while (options[++i].widgetname) {
    if (strcmp(widgetname, options[i].widgetname) != 0)
      continue;
    
    switch (options[i].type) {
    case TOGGLE:
      *(gint *)options[i].value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
      return;
      break;
    
    case ENTRY:
      *(const gchar **)options[i].value = (const gchar *)gtk_entry_get_text(GTK_ENTRY(widget));
      return;
      break;
    
    case OPTIONMENU:
      *(gint *)options[i].value = gtk_option_menu_get_history(GTK_OPTION_MENU(widget));
      return;
      break;
    
    default:
      g_assert_not_reached();
      break;
    }
  }
  
  fprintf(stderr, "Fatal Error: on_gui_addfiles_rule_widget_changed():\n Widget not found: %s\n", widgetname);
}
/***************************************************************************************
 * END CALLBACKS
 ***************************************************************************************/


void gui_addfiles_create(void)
{
  GtkWidget *treeview = NULL;
  int       i         = -1;
  
  /* Build the mainwindow. */
  if (!(gui.addfiles.xml = glade_xml_new(PACKAGE_DATA_DIR "/cantus-2.glade", "addfiles", NULL))) {
    fprintf(stderr, "Error: gui_init(): Interface file not found. (%s)", PACKAGE_DATA_DIR "/cantus.glade");
    exit(1);
  }
  
  glade_xml_signal_autoconnect(gui.addfiles.xml);
  gui.addfiles.widget = glade_xml_get_widget(gui.addfiles.xml, "addfiles");
  
  /* Create the filelist. */
  treeview = glade_xml_get_widget(gui.addfiles.xml, "treeview_addfiles_files_files");
  gui.addfiles.files_filelist_store = filelist_create(treeview);
  filelist_update(gui.addfiles.files_filelist_store, gui.defaultdir, cfg.file_pattern, gui.showhidden);
  
  /* Create the dirtree for the "addfiles" widget's "Files" tab. */
  treeview = glade_xml_get_widget(gui.addfiles.xml, "treeview_addfiles_files_directories");
  gui.addfiles.files_dirtree_store = dirtree_create(treeview);
  dirtree_open_directory(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(gui.addfiles.files_dirtree_store), gui.defaultdir);
  
  /* Create the dirtree for the "addfiles" widget's "Directories" tab. */
  treeview = glade_xml_get_widget(gui.addfiles.xml, "treeview_addfiles_directories_directories");
  gui.addfiles.directories_dirtree_store = dirtree_create(treeview);
  dirtree_open_directory(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(gui.addfiles.directories_dirtree_store), gui.defaultdir);
  
  /* Get the current values from the new-built GUI. */
  while (options[++i].widgetname) {
    switch (options[i].type) {
    case TOGGLE:
      *(gint *)options[i].value = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(glade_xml_get_widget(gui.addfiles.xml, options[i].widgetname)));
      break;
    
    case ENTRY:
      *(const gchar **)options[i].value = (const gchar *)gtk_entry_get_text(
        GTK_ENTRY(glade_xml_get_widget(gui.addfiles.xml, options[i].widgetname)));
      break;
    
    case OPTIONMENU:
      *(gint *)options[i].value = gtk_option_menu_get_history(
        GTK_OPTION_MENU(glade_xml_get_widget(gui.addfiles.xml, options[i].widgetname)));
      break;
    
    default:
      g_assert_not_reached();
      break;
    }
  }
}


void gui_addfiles_destroy(void)
{
  dirtree_destroy(gui.addfiles.files_dirtree_store);
  filelist_destroy(gui.addfiles.files_filelist_store);
  dirtree_destroy(gui.addfiles.directories_dirtree_store);
}
