/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <grp.h>
#include <pwd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <libintl.h>
#include "gui_lib_filelist.h"
#include "lib_shellpattern.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

enum COLUMNS {
  COL_FILENAME,
  COL_SIZE,
  COL_MODE,
  COL_OWNER,
  COL_PATH,
  N_COLUMNS
};

typedef struct {
  int col;
  gchar *name;
} Pair;

static Pair colmap[] = {
  { COL_FILENAME, N_("Filename") },
  { COL_SIZE,     N_("Size") },
  { COL_MODE,     N_("Mode") },
  { COL_OWNER,    N_("Owner") },
  { 0, NULL }
};


/******************************************************************************
 * STATICS
 ******************************************************************************/
/*
 * Returns TRUE if *path is a directory, FALSE if not.
 */
static gboolean is_dir(gchar *path)
{
  struct stat pstat;
  
  if (stat(path, &pstat) < 0)
    return(FALSE);
  
  return(S_ISDIR(pstat.st_mode) != 0);
}


/*
 * Self-explanatory.
 */
static gint get_file_size(gchar *filename)
{
  struct stat pstat;
  lstat(filename, &pstat);
  return (gint)pstat.st_size;
}


/*
 * Compare function for tree sorting.
 */
static gint compare_alphabetically_case_insensitive(GtkTreeModel *model,
           GtkTreeIter  *a,
           GtkTreeIter  *b,
           gpointer      trash)
{
  gchar *dir1 = NULL, *dir2 = NULL;
  int col = 0;
  GtkSortType order;
  
  gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(model), &col, &order);
  gtk_tree_model_get(model, a, col, &dir1, -1);
  gtk_tree_model_get(model, b, col, &dir2, -1);
  
  return (strcasecmp(dir1, dir2));
}
/******************************************************************************
 * END STATICS
 ******************************************************************************/


/*
 * Initialize a treeview widget as directory tree and add the rootdir.
 * Arguments: treeview: The treeview widget to put this rootdir in.
 */
GtkListStore *filelist_create(GtkWidget *treeview)
{
  GtkListStore *store = NULL;
  GtkCellRenderer *renderer = NULL;
  GtkTreeViewColumn *column = NULL;
  int i = 0;
  
  g_assert(treeview != NULL);
  
  /* Create a model and connect signals. */
  store = gtk_list_store_new(N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(store));
//  gtk_signal_connect(GTK_OBJECT(treeview), "drag-drop", (GtkSignalFunc)on_filelist_drag_drop, NULL);
  
  /* Create a cell render. */
  renderer = gtk_cell_renderer_text_new();
  
  /* Add the columns to the view. */
  for (i = 0; i < N_COLUMNS - 1; i++) {
    column = gtk_tree_view_column_new_with_attributes(_(colmap[i].name), renderer, "text", i, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_column_set_sort_column_id(column, i);
  }
  
  /* Set sort column. */
  gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store), COL_FILENAME, (GtkTreeIterCompareFunc)compare_alphabetically_case_insensitive, NULL, NULL);
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(store), COL_FILENAME, GTK_SORT_ASCENDING);
  
  /* Selection type. */
  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)), GTK_SELECTION_MULTIPLE);
  
  return(store);
}


/*
 * Update the filelist.
 */
void filelist_update(GtkListStore *store, const gchar *directory, const gchar *pattern[], gboolean showhidden)
{
  GtkTreeIter   iter;
  DIR           *stream       = NULL;
  struct dirent *filestruct   = NULL;
  struct stat   filestatus;
  struct passwd *pwd;
  struct group  *grp;
  gchar         size[20], mode[30];
  gchar         *owner        = NULL;
  gchar         fullfilename[4096];
  gchar         *uname        = NULL;
  gchar         *fname        = NULL;
  gchar         *ffname       = NULL;
  
  gtk_list_store_clear(store);
  
  /* Open the directory. */
  if (!(stream = opendir(directory))) {
    printf("Error: filelist_update(): Failed opening path. (%s)\n", directory);
    return;
  }
  
  /* Add each child-dir to the node. */
  while ((filestruct = readdir(stream)) != NULL) {
    /* Show hidden files? */
    if (strncmp(filestruct->d_name, ".", 1) == 0 && !showhidden)
      continue;
    
    /* Does it match the pattern? */
    if (pattern && !shellpattern_match_any(filestruct->d_name, pattern, FALSE))
      continue;
    
    /* Create a fullpath. */
    snprintf(fullfilename, 4095, "%s%s", directory, filestruct->d_name);
    
    if (is_dir(fullfilename))
      continue;
    
    if (stat(fullfilename, &filestatus) != 0)
      continue;
    
    snprintf(size, 19, "%i kb", get_file_size(fullfilename) / 1024);
    snprintf(mode, 29, "%i%i%i", (gint)(filestatus.st_mode & 0700) >> 6, (gint)(filestatus.st_mode & 0070) >> 3, (gint)(filestatus.st_mode & 0007));
    
    pwd = getpwuid(filestatus.st_uid);
    grp = getgrgid(filestatus.st_gid);
    if (pwd && grp)
      owner = g_strconcat(pwd->pw_name ? pwd->pw_name : "?", ":", grp->gr_name ? grp->gr_name : "?", NULL);
    else if (pwd)
      owner = g_strconcat(pwd->pw_name ? pwd->pw_name : "?", ":?", NULL);
    else if (grp)
      owner = g_strconcat("?:", grp->gr_name ? grp->gr_name : "?", NULL);
    else
      owner = g_strdup("");
    
    /* Add the new node. */
    gtk_list_store_append(store, &iter);
    
    fname   = g_locale_to_utf8(filestruct->d_name, -1, NULL, NULL, NULL);
    uname   = g_locale_to_utf8(owner, -1, NULL, NULL, NULL);
    ffname  = g_locale_to_utf8(fullfilename, -1, NULL, NULL, NULL);
    
    gtk_list_store_set(store, &iter,
        COL_FILENAME, fname,
        COL_SIZE, size,
        COL_MODE, mode,
        COL_OWNER, uname,
        COL_PATH, ffname,
        -1);
    
    g_free(fname);
    g_free(uname);
    g_free(ffname);
    g_free(owner);
  }
  
  closedir(stream);
}


void filelist_selection_for_each(GtkTreeView *treeview, void (* func)(const char *filename))
{
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GList *sellist = NULL;
  GtkTreeIter iter;
  GList *item = NULL;
  const gchar *filename = NULL;
  
  g_assert(treeview != NULL);
  
  if (!(selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview))))
    return;
  
  if (!(sellist = gtk_tree_selection_get_selected_rows(selection, &model)))
    return;
  g_assert(model != NULL);
  
  item = sellist;
  while (item) {
    /* Get the filename of the given gtk path. */
    if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)item->data)) {
      fprintf(stderr, "Error: filelist_selection_for_each(): Failed receiving the iter.");
      return;
    }
    gtk_tree_model_get(model, &iter, COL_PATH, &filename, -1);
    
    /* Call the user function. */
    func(filename);
    
    item = item->next;
  }
  
  g_list_foreach(sellist, (GFunc)gtk_tree_path_free, NULL);
  g_list_free(sellist);
}


void filelist_destroy(GtkListStore *model)
{
  GtkTreeIter iter;
  
  g_return_if_fail(model != NULL);
  
  gtk_list_store_clear(model);
}

#undef COLUMNS
