/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <libintl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "gui_preferences.h"
#include "gui_init.h"
#include "engine_init.h"
#include "lib_configfile.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern Cfg cfg;
extern GUI gui;
extern Engine engine;
static gchar configfilename[2048];

enum COLUMNS {
  GROUP_COLUMN,
  N_COLUMNS
};

gchar *categories[] = {
  "Cantus",
  N_("General"),
  N_("Tagging"),
  N_("Preview Popups"),
  NULL
};

typedef struct {
  gchar *category;
  gchar *option;
  gchar *widgetname;
  gint type;
  void *value;
} Map;

static Map options[] = {
  /* Invisible options. */
  { "program internal", "cfgversion", NULL, INTEGER, &cfg.compat },
  
  /* General options. */
  { "program general", "show hidden files", "check_preferences_general_showhidden", INTEGER, &gui.showhidden },
  { "program general", "default dir", "entry_preferences_general_defaultdir", STRING, gui.defaultdir },
  { "program general", "save rulesets on exit", "check_preferences_general_saverulesets", INTEGER, &cfg.save_rulesets_on_exit },
  
  /* Tagging options. */
  { "program tagging", "tag encoding", "check_preferences_tagging_unicode", INTEGER, &cfg.use_unicode },
  { "program tagging", "default tag", "radio_preferences_tagging_defaulttag_v2", INTEGER, &cfg.defaulttag_v2 },
  { "program preview", "tageditor", "check_preferences_tagging_tageditor_preview", INTEGER, &cfg.preview_tageditor },
  { "program preview", "filename2tag", "check_preferences_tagging_file2tag_preview", INTEGER, &cfg.preview_file2tag },
  { "program preview", "masseditor", "check_preferences_tagging_masseditor_preview", INTEGER, &cfg.preview_masseditor },
  
  { NULL, NULL, NULL, 0, NULL }
};


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static void set_configfilename(void)
{
  snprintf(configfilename, 2047, "%s/.cantus2rc", getenv("HOME"));
}


/*
 * Build the Glade interface.
 */
static void gui_preferences_build(void)
{
  /* Build the window. */
  if (!(gui.preferences_xml = glade_xml_new(PACKAGE_DATA_DIR "/cantus-2.glade", "preferences", NULL))) {
    fprintf(stderr, "Error: gui_preferences_build(): Interface file not found. (%s)", PACKAGE_DATA_DIR "/cantus.glade");
    exit(1);
  }
  g_assert(gui.preferences_xml != NULL);
  
  gui.preferences = glade_xml_get_widget(gui.preferences_xml, "preferences");
  g_assert(gui.preferences != NULL);
  
  glade_xml_signal_autoconnect(gui.preferences_xml);
}


/*
 * Change the title label of the current preferences group.
 */
static void preferences_title_set(const char *text)
{
  GtkWidget *label = glade_xml_get_widget(gui.preferences_xml, "label_preferences_title");
  char title[2048];
  
  if (text)
    snprintf(title, 2047, "<big><b>%s</b></big>", text);
  else
    snprintf(title, 2047, "<big><b>Cantus %s</b></big>", VERSION);
  
  /* Change the title label. */
  gtk_label_set_text((GtkLabel *)label, title);
  gtk_label_set_use_markup((GtkLabel *)label, TRUE);
  gtk_widget_set_size_request(label, -1, 35);
  gtk_misc_set_alignment ((GtkMisc *)label, 0.97, 0.5);
  gtk_widget_show(label);
}


static void gui_preferences_selection_changed(GtkTreeSelection *selection, void *trash)
{
  GtkWidget     *notebook = glade_xml_get_widget(gui.preferences_xml, "notebook_preferences_main");
  GtkTreeView   *treeview = GTK_TREE_VIEW(glade_xml_get_widget(gui.preferences_xml, "treeview_preferences_group"));
  GtkTreeModel  *model    = gtk_tree_view_get_model(GTK_TREE_VIEW(treeview));
  GtkTreePath   *path     = NULL;
  GtkTreeIter   iter;
  char          *grpname  = NULL;
  int           i         = 0;
  
  g_assert(notebook != NULL);
  g_assert(model != NULL);
  
  if (!(selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview))))
    return;
  
  if (!gtk_tree_selection_get_selected_rows(selection, &model))
    return;
  
  path = (GtkTreePath *)gtk_tree_selection_get_selected_rows(selection, &model)->data;
  g_assert(path != NULL);
  
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_tree_model_get(model, &iter, GROUP_COLUMN, &grpname, -1);
  
  do {
    if (strcmp(_(categories[i]), _(grpname)) == 0) {
      gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), i);
      if (i == 0)
        preferences_title_set(NULL);
      else
        preferences_title_set(_(grpname));
      g_free(grpname);
      return;
    }
  } while(categories[++i] != NULL);
}


static void preferences_options_fill(void)
{
  GtkWidget *widget = NULL;
  int i = -1;
  
  while (options[++i].option) {
    if (!options[i].widgetname)
      continue;
    
    widget = glade_xml_get_widget(gui.preferences_xml, options[i].widgetname);
    
    if (strncmp(options[i].widgetname, "check" , 5) == 0) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), (gboolean)*(int *)options[i].value);
      continue;
    }
    
    if (strncmp(options[i].widgetname, "radio" , 5) == 0) {
      if (*(int *)options[i].value != 0)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), TRUE);
      continue;
    }
    
    if (strncmp(options[i].widgetname, "entry" , 5) == 0) {
      gtk_entry_set_text(GTK_ENTRY(widget), (gchar *)options[i].value);
      continue;
    }
    
    g_assert_not_reached();
  }
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/***************************************************************************************
 * CALLBACKS
 ***************************************************************************************/
void on_button_preferences_close_clicked(void)
{
  preferences_save();
  gtk_widget_hide(gui.preferences);
}


int on_preferences_delete_event(void)
{
  on_button_preferences_close_clicked();
  return(TRUE);
}


void on_check_preferences_general_showhidden_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  gui.showhidden = gtk_toggle_button_get_active(togglebutton);
}


void on_entry_preferences_general_defaultdir_changed(GtkEntry *entry, gpointer trash)
{
  strncpy(gui.defaultdir, gtk_entry_get_text(entry), 2047);
  if (*(gui.defaultdir + strlen(gui.defaultdir) - 1) != '/')
    snprintf(gui.defaultdir, 2047, "%s/", gtk_entry_get_text(entry));
}


void on_radio_preferences_tagging_defaulttag_v2_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  pthread_mutex_lock(&engine.mut);
  cfg.defaulttag_v2 = gtk_toggle_button_get_active(togglebutton);
  pthread_mutex_unlock(&engine.mut);
}


void on_check_preferences_tagging_unicode_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  pthread_mutex_lock(&engine.mut);
  cfg.use_unicode = gtk_toggle_button_get_active(togglebutton);
  pthread_mutex_unlock(&engine.mut);
}


void on_check_preferences_tagging_tageditor_preview_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  pthread_mutex_lock(&engine.mut);
  cfg.preview_tageditor = gtk_toggle_button_get_active(togglebutton);
  pthread_mutex_unlock(&engine.mut);
}


void on_check_preferences_tagging_masseditor_preview_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  pthread_mutex_lock(&engine.mut);
  cfg.preview_masseditor = gtk_toggle_button_get_active(togglebutton);
  pthread_mutex_unlock(&engine.mut);
}


void on_check_preferences_general_saverulesets_toggled(GtkToggleButton *togglebutton, gpointer trash)
{
  pthread_mutex_lock(&engine.mut);
  cfg.save_rulesets_on_exit = gtk_toggle_button_get_active(togglebutton);
  pthread_mutex_unlock(&engine.mut);
}
/***************************************************************************************
 * CALLBACKS
 ***************************************************************************************/


void gui_preferences_init(void)
{
  GtkTreeStore *store = NULL;
  GtkWidget *treeview = NULL;
  GtkTreeIter parentiter;
  GtkTreeViewColumn *column = NULL;
  GtkCellRenderer *renderer = NULL;
  int i = 0;
  
  if (!gui.preferences)
    gui_preferences_build();
  
  treeview = glade_xml_get_widget(gui.preferences_xml, "treeview_preferences_group");
  g_assert(treeview != NULL);
  
  /* Create a model.  We are using the store model for now, though we
  * could use any other GtkTreeModel */
  store = gtk_tree_store_new(N_COLUMNS, G_TYPE_STRING);
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(store));
  
  /* Create a cell render and arbitrarily make it red for demonstration purposes */
  renderer = gtk_cell_renderer_text_new();
  
  /* Create a column, associating the "text" attribute of the
  * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes(_("Groups"), renderer, "text", GROUP_COLUMN, NULL);
  
  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
  
  /* Add the groups to the treeview. */
  i = -1;
  while (categories[++i]) {
    gtk_tree_store_append(store, &parentiter, NULL);
    gtk_tree_store_set(store, &parentiter, GROUP_COLUMN, _(categories[i]), -1);
  }
  
  /* Selection type & callback. */
  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)), GTK_SELECTION_MULTIPLE);
  g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview))), "changed", G_CALLBACK(gui_preferences_selection_changed), NULL);
  
  gtk_tree_view_expand_all(GTK_TREE_VIEW(treeview));
  
  /* Select the root node. */
  gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &parentiter);
  gtk_tree_selection_select_iter(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)), &parentiter);
  
  /* Fill the GUI with the options. */
  preferences_options_fill();
}


void gui_preferences_activate_tab(int tab)
{
  GtkWidget *notebook = glade_xml_get_widget(gui.preferences_xml, "notebook_preferences_main");
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  
  gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), tab);
  if (tab == 0)
    preferences_title_set(NULL);
  else
    preferences_title_set(_(categories[tab]));
  
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(glade_xml_get_widget(gui.preferences_xml, "treeview_preferences_group")));
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(glade_xml_get_widget(gui.preferences_xml, "treeview_preferences_group")));
  gtk_tree_model_iter_nth_child(model, &iter, NULL, tab);
  gtk_tree_selection_unselect_all(selection);
  gtk_tree_selection_select_iter(selection, &iter);
}


void gui_preferences_destroy(void)
{
  gtk_widget_destroy(gui.preferences);
}


void preferences_save(void)
{
  gchar config[CONFIGFILE_MAXSIZE + 1];
  int i = -1;
  
  set_configfilename();
  configfile_read(configfilename, config, CONFIGFILE_MAXSIZE);
  
  while (options[++i].option) {
    pthread_mutex_lock(&engine.mut);
    switch (options[i].type) {
    case INTEGER:
      configfile_option_add_integer(config, options[i].category, options[i].option, options[i].value);
      break;
    
    case STRING:
      configfile_option_add(config, options[i].category, options[i].option, options[i].value);
      break;
    
    default:
      g_assert_not_reached();
    }
    pthread_mutex_unlock(&engine.mut);
  }
  
  configfile_write(configfilename, config);
}


void preferences_load(void)
{
  gchar config[CONFIGFILE_MAXSIZE + 1];
  int i = -1;
  
  set_configfilename();
  if (configfile_read(configfilename, config, CONFIGFILE_MAXSIZE) < 0) {
    gui_popup(gui.mainwindow, _("Welcome To Cantus 2!\n\n"
      "This seems to be the first time you are starting Cantus 2.\n\n"
      "* If you find any bugs, please report them\n"
      "  at http://www.debain.org\n"
      "* Translators wanted! Support for new languages\n"
      "  will be added to the Cantus 2 package."),
      GTK_MESSAGE_INFO);
    preferences_save();
    return;
  }
  
  while (options[++i].option) {
    pthread_mutex_lock(&engine.mut);
    switch (options[i].type) {
    case INTEGER:
      configfile_option_get_integer(config, options[i].category, options[i].option, options[i].value);
      break;
    
    case STRING:
      configfile_option_get(config, options[i].category, options[i].option, options[i].value);
      break;
    
    default:
      g_assert_not_reached();
    }
    pthread_mutex_unlock(&engine.mut);
  }
}

#undef COLUMNS
