/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <libintl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "engine_init.h"
#include "engine_fileinfo.h"
#include "engine_queue.h"
#include "gui_init.h"
#include "gui_tageditor.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern Cfg cfg;
extern Engine engine;
extern GUI gui;

enum COLUMNS {
  COL_DIRECTORY,
  COL_FILENAME,
  COL_TAGTYPE,
  COL_ARTIST,
  COL_ALBUM,
  COL_SONG,
  COL_TRACK,
  COL_YEAR,
  COL_GENRE,
  COL_DATA,
  N_COLUMNS
};

typedef struct {
  int   col;
  gchar *name;
  int   markup;
} Pair;

static Pair colmap[] = {
  { COL_DIRECTORY,  N_("Directory"),  0 },
  { COL_FILENAME,   N_("Filename"),   0 },
  { COL_TAGTYPE,    N_("Tag"),        1 },
  { COL_ARTIST,     N_("Artist"),     1 },
  { COL_ALBUM,      N_("Album"),      1 },
  { COL_SONG,       N_("Song"),       1 },
  { COL_TRACK,      N_("Track"),      1 },
  { COL_YEAR,       N_("Year"),       1 },
  { COL_GENRE,      N_("Genre"),      1 },
  { 0, NULL }
};


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
typedef struct {
  void *info;
  int column;
  GtkTreeIter *iter;
  gboolean found;
} ForEachCmpData;

static gboolean pointer_same(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ForEachCmpData *rowdata)
{
  void * data = NULL;
  
  gtk_tree_model_get(model, iter, rowdata->column, &data, -1);
  
  if (rowdata->info == data) {
    memcpy(rowdata->iter, iter, sizeof(GtkTreeIter));
    rowdata->found = TRUE;
    return(TRUE);
  }
  
  return(FALSE);
}


static gboolean queue_item_remove(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, GList **iterlist)
{
  FileInfo *info = NULL;
  
  /* Get the current GUI item. */
  gtk_tree_model_get(model, iter, COL_DATA, &info, -1);
  g_assert(info != NULL);
  
  /* If the item is still in the queue, do nothing. */
  if (engine.queue && g_hash_table_lookup(engine.queue, info))
    return(FALSE);
  
  /* Otherwise store the iter of this item in a list for later removal. */
  *iterlist = g_list_append(*iterlist, gtk_tree_iter_copy(iter));
  
  return(FALSE);
}


/*
 * Find the tree row/column that contains the file "info"
 * and store the position in "iter".
 * Returns TRUE on success, otherwise FALSE.
 */
static gboolean find_treeview_row(GtkTreeModel *model, GtkTreeIter * iter, int column, void * info)
{
  ForEachCmpData data;
  
  data.info = info;
  data.column = column;
  data.iter = iter;
  data.found = FALSE;
  
  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)pointer_same, &data);
  
  return(data.found);
}


static gchar *field_markup(gchar *text1, gchar *text2)
{
  gchar         *mark1  = NULL;
  gchar         *mark2  = NULL;
  gchar         *mark   = NULL;
  const gchar   *end;
  
  text1 = text1 ? text1 : "";
  text2 = text2 ? text2 : "";
  
  /* Be UTF-8 save. */
  if (!g_utf8_validate(text1, -1, &end)) {
    gchar *new = g_strndup(text1, g_utf8_pointer_to_offset(text1, end));
    g_free(text1);
    text1 = new;
  }
  if (!g_utf8_validate(text2, -1, &end)) {
    gchar *new = g_strndup(text2, g_utf8_pointer_to_offset(text2, end));
    g_free(text2);
    text2 = new;
  }
  
  mark1 = g_markup_escape_text(text1 ? text1 : "", -1);
  mark2 = g_markup_escape_text(text2 ? text2 : "", -1);
  mark  = g_strdup_printf("%s\n<span color='dim grey'>%s</span>", mark1, mark2);
  
  g_free(mark1);
  g_free(mark2);
  
  return(mark);
}


static GHFunc list_item_add(FileInfo *info)
{
  GtkListStore  *store        = gui.queue;
  GtkTreeIter   iter;
  gchar         *file         = NULL;
  gchar         *dir          = NULL;
  Tag           *tag[2];
  gchar         type[256];
  gchar         *artist       = NULL;
  gchar         *album        = NULL;
  gchar         *title        = NULL;
  gchar         *track        = NULL;
  gchar         *year         = NULL;
  gchar         *genre        = NULL;
  const gchar   *end;
  
  g_assert(info != NULL);
  
  if (!info->updated)
    return(NULL);
  
  /* Default tag order. */
  if (cfg.defaulttag_v2) {
    tag[0]  = info->tag[1];
    tag[1]  = info->tag[0];
  }
  else {
    tag[0]  = info->tag[0];
    tag[1]  = info->tag[1];
  }
  
  artist = field_markup(tag[0] ? tag[0]->artist : NULL, tag[1] ? tag[1]->artist : NULL);
  album  = field_markup(tag[0] ? tag[0]->album  : NULL, tag[1] ? tag[1]->album  : NULL);
  title  = field_markup(tag[0] ? tag[0]->title  : NULL, tag[1] ? tag[1]->title  : NULL);
  track  = field_markup(tag[0] ? tag[0]->track  : NULL, tag[1] ? tag[1]->track  : NULL);
  year   = field_markup(tag[0] ? tag[0]->year   : NULL, tag[1] ? tag[1]->year   : NULL);
  genre  = field_markup(tag[0] ? tag[0]->genre  : NULL, tag[1] ? tag[1]->genre  : NULL);
  
  switch (info->tag_type) {
  case ID3:
    if (cfg.defaulttag_v2)
      snprintf(type, 255, "%s\n<span color='dim grey'>%s</span>", tag[0] ? _("V2") : "", tag[1] ? _("V1") : "");
    else
      snprintf(type, 255, "%s\n<span color='dim grey'>%s</span>", tag[0] ? _("V1") : "", tag[1] ? _("V2") : "");
    break;
    
  case OGG:
    if (tag[0])
      strncpy(type, _("OGG\n"), 255);
    else
      *type = '\0';
    break;
  }
  
  /* Split the filename into file and dir name. */
  file = strdup(strrchr(info->filename, '/') + 1);
  dir  = strdup(info->filename);
  if (strrchr(dir, '/'))
    *(strrchr(dir, '/') + 1) = '\0';
  
  /* If it's not yet in the list, append it. */
  if (!find_treeview_row(GTK_TREE_MODEL(store), &iter, COL_DATA, info))
    gtk_list_store_append(store, &iter);
  
  /* Be UTF-8 save. */
  if (!g_utf8_validate(file, -1, &end)) {
    gchar *new = g_strndup(file, g_utf8_pointer_to_offset(file, end));
    g_free(file);
    file = new;
  }
  if (!g_utf8_validate(dir, -1, &end)) {
    gchar *new = g_strndup(dir, g_utf8_pointer_to_offset(dir, end));
    g_free(dir);
    dir = new;
  }
  
  gtk_list_store_set(store, &iter,
    COL_FILENAME,   file,
    COL_DIRECTORY,  dir,
    COL_TAGTYPE,    type,
    COL_ARTIST,     artist,
    COL_ALBUM,      album,
    COL_SONG,       title,
    COL_TRACK,      track,
    COL_YEAR,       year,
    COL_GENRE,      genre,
    COL_DATA, info, -1);
  
  info->updated = FALSE;
  
  g_free(artist);
  g_free(album);
  g_free(title);
  g_free(track);
  g_free(year);
  g_free(genre);
  
  return(NULL);
}


static void queue_selection_changed(GtkTreeSelection *selection, void *trash)
{
  gui_tageditor_clear();
  gui_tageditor_update();
  gui_file2tag_update();
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


/*
 * Initialize the queue GUI.
 */
GtkListStore *gui_queue_create(GtkTreeView *treeview)
{
  GtkListStore      *store    = NULL;
  GtkTreeViewColumn *column   = NULL;
  GtkCellRenderer   *renderer = NULL;
  int               i         = 0;

  g_assert(treeview != NULL);
  
  /* Create a model and connect signals. */
  store = gtk_list_store_new(N_COLUMNS,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_POINTER);
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(store));
//  gtk_signal_connect(GTK_OBJECT(treeview), "drag-drop", (GtkSignalFunc)on_queue_drag_drop, NULL);
  
  /* Add the columns to the view. */
  for (i = 0; i < N_COLUMNS - 1; i++) {
    /* Create a cell renderer. */
    renderer = gtk_cell_renderer_text_new();
    gtk_cell_renderer_text_set_fixed_height_from_font(GTK_CELL_RENDERER_TEXT(renderer), 2);
    
    /* Create the column. */
    column = gtk_tree_view_column_new_with_attributes(_(colmap[i].name), renderer, colmap[i].markup ? "markup" : "text", i, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_column_set_sort_column_id(column, i);
  }
  
  /* Selection type & callback. */
  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)), GTK_SELECTION_MULTIPLE);
  g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview))), "changed", G_CALLBACK(queue_selection_changed), NULL);
  
  return(store);
}


/*
 * Update the queue GUI.
 */
void gui_queue_update(GtkTreeModel *model)
{
  GtkTreeIter iter;
  GList       *iterlist = NULL;
  GList       *item     = NULL;
  
  g_assert(model != NULL);
  
  
  /* Remove dead items from the GUI queue. */
  if (gtk_tree_model_get_iter_first(model, &iter)) {
    gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)queue_item_remove, &iterlist);
    item = iterlist;
    while (item) {
      gtk_list_store_remove(GTK_LIST_STORE(model), (GtkTreeIter*)item->data);
      gtk_tree_iter_free((GtkTreeIter*)item->data);
      item = item->next;
    }
    g_list_free(iterlist);
  }
  
  /* If the queue is empty, we're done. */
  if (!engine.queue)
    return;
  
  /* Now add all new queue items to the GUI. */
  g_hash_table_foreach(engine.queue, (GHFunc)list_item_add, NULL);
}


FileInfo *gui_queue_row_info_get(GtkTreeView *treeview, GtkTreePath *path)
{
  FileInfo      *info   = NULL;
  GtkTreeModel  *model  = gtk_tree_view_get_model(treeview);
  GtkTreeIter   iter;
  
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, COL_DATA, &info, -1);
  
  return(info);
}


void gui_queue_selection_for_each(GtkTreeView *treeview, void (* func)(FileInfo *info, void *data), void *data)
{
  GtkTreeSelection  *selection  = NULL;
  GtkTreeModel      *model      = NULL;
  GList             *sellist    = NULL;
  GtkTreeIter       iter;
  GList             *item       = NULL;
  FileInfo          *info       = NULL;
  
  g_assert(treeview != NULL);
  
  if (!(selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview))))
    return;
  
  if (!(sellist = gtk_tree_selection_get_selected_rows(selection, &model)))
    return;
  g_assert(model != NULL);
  
  item = sellist;
  while (item) {
    /* Get the filename of the given gtk path. */
    if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)item->data)) {
      fprintf(stderr, "Error: filelist_selection_for_each(): Failed receiving the iter.");
      return;
    }
    gtk_tree_model_get(model, &iter, COL_DATA, &info, -1);
    
    /* Call the user function. */
    func(info, data);
    
    item = item->next;
  }
  
  g_list_foreach(sellist, (GFunc)gtk_tree_path_free, NULL);
  g_list_free(sellist);
}


/*
 * Returns the data of the first selected row in the queue, or NULL.
 */
void *gui_queue_get_selection_first_data(GtkTreeView *treeview)
{
  GtkTreeSelection  *selection  = NULL;
  GtkTreeModel      *model      = NULL;
  GList             *sellist    = NULL;
  GtkTreeIter       iter;
  void              *data;
  
  /* Receive the model. */
  if (!(selection = gtk_tree_view_get_selection(treeview)))
    return;
  if (!(sellist = gtk_tree_selection_get_selected_rows(selection, &model)))
    return;
  g_assert(model != NULL);
  
  /* Receive the iter. */
  if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)sellist->data)) {
    fprintf(stderr, "Error: filelist_selection_for_each(): Failed receiving the iter.");
    return;
  }
  gtk_tree_model_get(model, &iter, COL_DATA, &data, -1);
  
  return(data);
}

#undef COLUMNS
