/* The Cantus project.
 * (c)2002, 2003 by Samuel Abels (spam debain org)
 * This project's homepage is: http://www.debain.org/cantus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <libintl.h>
#include <pthread.h>
#include <stdio.h>
#include <glade/glade.h>
#include "gui_statementlist.h"
#include "gui_init.h"
#include "engine_init.h"
#include "engine_ruleset.h"

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)

extern GUI    gui;
extern Engine engine;

enum COLUMNS {
  COL_COND,
  COL_RULE,
  COL_FIELD,
  COL_DATA,
  N_COLUMNS
};

typedef struct {
  int   col;
  gchar *name;
} Pair;

static Pair colmap[] = {
  { COL_COND, N_("Condition") },
  { COL_RULE, N_("Rule") },
  { COL_RULE, N_("Field") },
  { 0, NULL }
};

typedef struct {
  void *info;
  int column;
  GtkTreeIter *iter;
  gboolean found;
} ForEachCmpData;

typedef struct {
  GList *statements;
  GList **iterlist;
} TreeModelForEachArg;


/***************************************************************************************
 * STATICS
 ***************************************************************************************/
static gboolean pointer_same(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, ForEachCmpData *rowdata)
{
  void *data = NULL;
  
  gtk_tree_model_get(model, iter, rowdata->column, &data, -1);
  
  if (rowdata->info == data) {
    memcpy(rowdata->iter, iter, sizeof(GtkTreeIter));
    rowdata->found = TRUE;
    return(TRUE);
  }
  
  return(FALSE);
}


static gboolean gui_statementlist_item_remove(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, TreeModelForEachArg *arg)
{
  RuleStatement *statement  = NULL;
  GList         *item       = NULL;
  
  /* Get the current GUI item. */
  gtk_tree_model_get(model, iter, COL_DATA, &statement, -1);
  g_assert(statement != NULL);
  
  /* If the item is still in the queue, do nothing. */
  item = arg->statements;
  while (item) {
    if (statement == item->data)
      return(FALSE);
    item = item->next;
  }
  
  /* Otherwise store the iter of this item in a list for later removal. */
  *(arg->iterlist) = g_list_append(*(arg->iterlist), gtk_tree_iter_copy(iter));
  
  return(FALSE);
}


/*
 * Find the tree row/column that contains the file "info"
 * and store the position in "iter".
 * Returns TRUE on success, otherwise FALSE.
 */
static gboolean find_treeview_row(GtkTreeModel *model, GtkTreeIter *iter, int column, void *info)
{
  ForEachCmpData data;
  
  data.info   = info;
  data.column = column;
  data.iter   = iter;
  data.found  = FALSE;
  
  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)pointer_same, &data);
  
  return(data.found);
}


static GFunc gui_statementlist_item_add(RuleStatement *statement, GtkListStore *store)
{
  GtkTreeIter iter;
  
  g_assert(statement != NULL);
  
  /* If it is not already in the list, append it. */
  if (!find_treeview_row(GTK_TREE_MODEL(store), &iter, COL_DATA, statement))
    gtk_list_store_append(store, &iter);
  
  gtk_list_store_set(store, &iter,
    COL_COND,   statement->cond.descr,
    COL_RULE,   statement->rule.descr,
    COL_FIELD,  statement->dest_descr,
    COL_DATA,   statement, -1);
  
  return(NULL);
}


static gboolean on_statementlist_drag_drop(GtkWidget *widget, GdkDragContext *ct, gint a, gint b, guint c, StamentlistDragDropFunc func)
{
  if (func)
    func((GtkTreeView*)widget);
  return(FALSE);
}
/***************************************************************************************
 * END STATICS
 ***************************************************************************************/


void gui_statementlist_init(GtkTreeView *treeview, StamentlistDragDropFunc func)
{
  GtkListStore      *store    = NULL;
  GtkTreeViewColumn *column   = NULL;
  GtkCellRenderer   *renderer = NULL;
  gint              i         = 0;
  
  g_assert(treeview != NULL);
  
  /* Create a model and connect signals. */
  store = gtk_list_store_new(N_COLUMNS,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_STRING,
    G_TYPE_POINTER);
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(store));
  gtk_signal_connect_after(GTK_OBJECT(treeview), "drag-drop", (GtkSignalFunc)on_statementlist_drag_drop, func);
  
  /* Create a cell render. */
  renderer = gtk_cell_renderer_text_new();
  
  /* Add the columns to the view. */
  for (i = 0; i < N_COLUMNS - 1; i++) {
    column = gtk_tree_view_column_new_with_attributes(_(colmap[i].name), renderer, "text", i, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
  }
  
  /* Selection type. */
  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview)), GTK_SELECTION_MULTIPLE);
}


void gui_statementlist_update(GtkListStore *store, Ruleset *ruleset)
{
  GtkTreeIter         iter;
  GList               *iterlist = NULL;
  GList               *item     = NULL;
  GtkTreeView         *tree     = NULL;
  TreeModelForEachArg arg;
  
  g_assert(store != NULL);
  
  if (ruleset)
    arg.statements = ruleset->statements;
  else
    arg.statements = NULL;
  arg.iterlist = &iterlist;
  
  /* Lock everything. */
  pthread_mutex_lock(&engine.mut);
  
  /* Remove dead items from the list. */
  if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter)) {
    gtk_tree_model_foreach(GTK_TREE_MODEL(store), (GtkTreeModelForeachFunc)gui_statementlist_item_remove, &arg);
    item = iterlist;
    while (item) {
      gtk_list_store_remove(store, (GtkTreeIter*)item->data);
      gtk_tree_iter_free((GtkTreeIter*)item->data);
      item = item->next;
    }
    g_list_free(iterlist);
  }
  
  /* If the list is empty, we're done. */
  if (!ruleset || !ruleset->statements) {
    pthread_mutex_unlock(&engine.mut);
    return;
  }
  
  /* Now add all new queue items to the GUI. */
  g_list_foreach(ruleset->statements, (GFunc)gui_statementlist_item_add, store);
  
  /* Unlock stuff. */
  pthread_mutex_unlock(&engine.mut);
}


/*
 * Receive the statement that is applied to the give statementlist row.
 */
void gui_statementlist_data_get_from_iter(GtkListStore *store, GtkTreeIter *iter, RuleStatement **sm)
{
  /* Get the current GUI item. */
  gtk_tree_model_get(GTK_TREE_MODEL(store), iter, COL_DATA, sm, -1);
}


/*
 * Move all selected nodes up by one.
 */
void gui_statementlist_selection_up(GtkTreeView *tree)
{
  GtkListStore      *store     = GTK_LIST_STORE(gtk_tree_view_get_model(tree));
  GtkTreeModel      *model     = GTK_TREE_MODEL(store);
  GtkTreeSelection  *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  GList             *sellist   = NULL;
  GList             *item      = NULL;
  GtkTreeIter       iter, previter;
  
  /* Get a path list of all selected items in the statement list. */
  if (!selection)
    return;
  sellist = gtk_tree_selection_get_selected_rows(selection, &model);
  
  /* Walk through all selected items. */
  while (sellist) {
    /* Get the current and the previous iter. */
    if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)sellist->data))
      break;
    if (!gtk_tree_path_prev((GtkTreePath*)sellist->data)
      || !gtk_tree_model_get_iter(model, &previter, (GtkTreePath*)sellist->data))
      break;
    
    /* Swap the rows. */
    gtk_list_store_swap(store, &iter, &previter);
    
    sellist = sellist->next;
  }
}

/*
 * Move all selected nodes up by one.
 */
void gui_statementlist_selection_down(GtkTreeView *tree)
{
  GtkListStore      *store     = GTK_LIST_STORE(gtk_tree_view_get_model(tree));
  GtkTreeModel      *model     = GTK_TREE_MODEL(store);
  GtkTreeSelection  *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
  GList             *sellist   = NULL;
  GList             *item      = NULL;
  GtkTreeIter       iter, nextiter;
  
  /* Get a *reversed* path list of all selected items in the statement list. */
  if (!selection)
    return;
  sellist = gtk_tree_selection_get_selected_rows(selection, &model);
  sellist = g_list_reverse(sellist);
  
  /* Walk through all selected items. */
  while (sellist) {
    /* Get the current and the next iter. */
    if (!gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)sellist->data))
      break;
    gtk_tree_path_next((GtkTreePath*)sellist->data);
    if (!gtk_tree_model_get_iter(model, &nextiter, (GtkTreePath*)sellist->data))
      break;
    
    /* Swap the rows. */
    gtk_list_store_swap(store, &iter, &nextiter);
    
    sellist = sellist->next;
  }
}


/*
 * Receive a doubly linked list that contains all statements.
 * This is required to apply the reordered list to the engine.
 */
GList *gui_statementlist_order_get(GtkListStore *store)
{
  RuleStatement *sm     = NULL;
  GList         *smlist = NULL;
  GtkTreeIter   iter;
  
  /* Walk through the GUI list. */
  if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter))
    return;
  do {
    /* Get the statement from the current row. */
    gui_statementlist_data_get_from_iter(store, &iter, &sm);
    if (!sm) {
      fprintf(stderr, "Fatal Error: gui_statementlist_order_get(): No data applied to the statement list row.\n");
      continue;
    }
    /* Put it in a list. */
    smlist = g_list_append(smlist, sm);
  } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter));
  
  return(smlist);
}

#undef COLUMNS
